# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/04_learner.ipynb (unless otherwise specified).

__all__ = ['style_learner']

# Cell
from fastai2.basics import *
from faststyle import *

# Cell
@delegates(Learner.__init__)
def style_learner(dls, model, get_feats, style_fns, loss_func=None, cbs=None, metrics=None, **kwargs):
  loss_func = loss_func or FastStyleLoss(get_feats)
  cbs = L(cbs) + L(FeatsCallback.from_fns(style_fns, get_feats))
  metrics = L(metrics) + getattr(loss_func, 'metrics', L())
  return Learner(dls, model, loss_func=loss_func, cbs=cbs, metrics=metrics, **kwargs)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/00_core.ipynb (unless otherwise specified).

__all__ = ['TEST_IMAGE', 'zip_safe', 'wmean', 'create']

# Cell
from fastai2.basics import *
from fastai2.vision.all import *

# Cell
TEST_IMAGE = 'images/puppy.jpg'

# Cell
def zip_safe(*args, **kwargs):
  'Raises ValueError is len of items are different. Does not work for generators'
  if len(set([len(o) for o in args])) != 1: raise ValueError('All items should have same length')
  return zip(*args, **kwargs)

# Cell
def _wmean(t:Tensor, w=None, dim=None):
  'Weighted mean'
  if w is None: return t.mean(dim=dim)
  w = tensor(w, device=t.device)
  assert w.sum() == 1., 'weights must sum to 1'
  if dim is not None:
    assert len(w) == t.shape[dim], 'weights must have the same number of items as chosen dimension'
    sz = torch.ones(len(t.shape), dtype=int)
    sz[dim] = t.shape[dim]
  else:
    assert len(w) == torch.prod(tensor(t.shape))
    sz = t.shape
  w = w.reshape(*sz)
  if dim is not None:  return (w*t).sum(dim=dim) #TODO: dim=None gives error
  else:                return (w*t).sum()

# Cell
@patch
@delegates(_wmean)
def wmean(self:Tensor, w, **kwargs): return _wmean(self, w, **kwargs)

# Cell
@patch_to(TensorImage)
def create(fn) -> TensorImage:
  pipe = Pipeline([PILImage.create, ToTensor(), IntToFloatTensor()])
  return pipe(fn).to(default_device())


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/02_layer_features.ipynb (unless otherwise specified).

__all__ = ['prepare_feat_model', 'get_layers', 'FeatModels', 'LayerFeats']

# Cell
from fastai2.basics import *
from fastai2.vision.all import *
from fastai2.callback.all import *
from torchvision.models import vgg16, vgg19
from faststyle import *

# Cell
def prepare_feat_model(m):
  m = m.to(default_device()).eval()
  for p in m.parameters(): p.requires_grad=False
  return m

# Cell
def get_layers(m, idxs):
  return [m[i] for i in idxs]

# Cell
_imagenet_norm = NormalizeAll.from_stats(*imagenet_stats)

# Cell
class FeatModels:
  @staticmethod
  def vgg16():
    m = prepare_feat_model(vgg16(True).features)
    stl_ls = get_layers(m, (1, 11, 18, 25))
    cnt_ls = get_layers(m, (20,))
    return dict(m=m, stl_ls=stl_ls, cnt_ls=cnt_ls, tfms=[_imagenet_norm])

  @staticmethod
  def vgg19():
    m = prepare_feat_model(vgg19(True).features)
    stl_ls = get_layers(m, (1, 6, 11, 20, 29))
    cnt_ls = get_layers(m, (22,))
    return dict(m=m, stl_ls=stl_ls, cnt_ls=cnt_ls, tfms=[_imagenet_norm])

# Cell
class LayerFeats(Module):
  def __init__(self, m, stl_ls, cnt_ls, tfms=None):
    self.m, self.tfms = m, Pipeline(tfms)
    self.stl_hooks = hook_outputs(stl_ls, detach=False)
    self.cnt_hooks = hook_outputs(cnt_ls, detach=False)

  def forward(self, x):
    _ = self.m(self.tfms(x))
    return self.stl_hooks.stored, self.cnt_hooks.stored

  @classmethod
  def from_feat_m(cls, feat_m): return cls(**feat_m())


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01_data.ipynb (unless otherwise specified).

__all__ = ['IncrementalSplitter', 'coco_stats', 'TensorImageX', 'PILImageX', 'NormalizeX', 'style_blocks',
           'NormalizeAll']

# Cell
from fastai2.basics import *
from fastai2.vision.all import *
from faststyle import *

# Cell
class IncrementalSplitter:
  'Dynamically changes pct of data used without mixing train and val data'
  @delegates(RandomSplitter)
  def __init__(self, pct=1., **kwargs):
    self.pct = pct
    seed = kwargs.pop('seed', None) or random.randint(0,1e6)
    self._splitter = RandomSplitter(seed=seed, **kwargs)

  def __call__(self, o, **kwargs):
    tidxs, vidxs = self._splitter(o, **kwargs)
    random.shuffle(tidxs); random.shuffle(vidxs)
    return tidxs[:int(len(tidxs)*self.pct)], vidxs[:int(len(vidxs)*self.pct)]

# Cell
coco_stats = ([0.4605, 0.4101, 0.3642], [0.2780, 0.2701, 0.2741])

# Cell
class TensorImageX(TensorImage): pass

# Cell
class PILImageX(PILImage): pass
PILImageX._tensor_cls = TensorImageX

# Cell
class NormalizeX(Normalize):
  def encodes(self, x:TensorImage, **kwargs): return x
  def decodes(self, x:TensorImage, **kwargs): return x

  def encodes(self, x:TensorImageX, **kwargs): return super().encodes(x, **kwargs)
  def decodes(self, x:TensorImageX, **kwargs): return super().decodes(x, **kwargs)

# Cell
style_blocks = ImageBlock(PILImageX), ImageBlock()

# Cell
class NormalizeAll(Normalize):
  def encodes(self, x:Tensor, **kwargs): return super().encodes[TensorImage](self, x, **kwargs)
  def decodes(self, x:Tensor, **kwargs): return super().decodes[TensorImage](self, x, **kwargs)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/06_callback.ipynb (unless otherwise specified).

__all__ = ['FeatsCallback']

# Cell
from fastai2.basics import *
from fastai2.callback.all import *
from faststyle import *

# Cell
# TODO: This allow loss functions to be separated
class FeatsCallback(Callback):
  def __init__(self, get_fts, stl_fts=None, cnt_fts=None):
    store_attr(self, 'get_fts,stl_fts,cnt_fts')
    # Change dict with a class. Can be accessed with .
    self.fts = dict(pred={}, targ={}, source={})
    self.fts['source']['stl'],self.fts['source']['cnt'] = L(stl_fts),L(cnt_fts)

  def after_pred(self):
    assert len(self.yb)==1
    fts = self.fts
    fts['pred']['stl'],fts['pred']['cnt'] = self.get_fts(self.pred)
    fts['targ']['stl'],fts['targ']['cnt'] = self.get_fts(self.yb[0])
    self.learn.yb = (*self.yb, fts)

  @classmethod
  def from_fns(cls, fns, get_fts):
    source_tims = L(TensorImage.create(fn) for fn in L(fns))
    stl_fts,cnt_fts = zip(*L(get_fts(o[None]) for o in source_tims))
    stl_fts,cnt_fts = L(zip(*stl_fts)),L(zip(*cnt_fts))
    return cls(get_fts=get_fts, stl_fts=stl_fts, cnt_fts=cnt_fts)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/03_loss.ipynb (unless otherwise specified).

__all__ = ['gram', 'ContentLoss', 'content_loss', 'StyleLoss', 'TVLoss', 'PixelLoss', 'pixel_loss', 'FastStyleLoss',
           'SuperResLoss']

# Cell
from fastai2.basics import *
from faststyle import *

# Cell
def gram(x):
    n,c,h,w = x.size()
    x = x.view(n, c, -1)
    return (x @ x.transpose(1,2))/(c*h*w)

# Cell
def ContentLoss():
  def _inner(pred, targ, fts):
    return sum([F.mse_loss(*o) for o in zip_safe(fts['pred']['cnt'],fts['targ']['cnt'])])
  return _inner
def content_loss(pred, targ, fts):
  return sum([F.mse_loss(*o) for o in zip_safe(fts['pred']['cnt'],fts['targ']['cnt'])])

# Cell
def StyleLoss(ws=None):
  def _inner(pred, targ, fts):
    bs = fts['pred']['stl'][0].shape[0]
    sims_gs = L(torch.stack([gram(f) for f in fs]).wmean(ws, dim=0) for fs in fts['source']['stl'])
    pred_gs = [gram(f) for f in fts['pred']['stl']]
    assert len(sims_gs) == len(pred_gs)
    stl_losses = [F.mse_loss(g1.repeat(bs,1,1),g2) for g1,g2 in zip(sims_gs,pred_gs)]
    return sum(stl_losses)
  return _inner

# Cell
def TVLoss():
  def _inner(pred,*_):
    tv_h = ((pred[:,:,1:,:] - pred[:,:,:-1,:]).pow(2)).mean()
    tv_w = ((pred[:,:,:,1:] - pred[:,:,:,:-1]).pow(2)).mean()
    return (tv_h + tv_w)
  return _inner

# Cell
def PixelLoss(base_loss=F.mse_loss):
  def _inner(pred, targ, fts, **kwargs):
    return base_loss(pred,targ,**kwargs)
def pixel_loss(pred, targ, fts, base_loss=F.mse_loss, **kwargs):
  return base_loss(pred,targ,**kwargs)

# Cell
#TODO: Use @func_kwargs for losses
class FastStyleLoss(Module):
  def __init__(self, get_fs, stl_w=3e5, cnt_w=1, tv_w=20):
    self.stl_loss_fn,self.cnt_loss_fn,self.tv_loss_fn = StyleLoss(),ContentLoss(),TVLoss()
    store_attr(self, 'get_fs,stl_w,cnt_w,tv_w')
    self.metric_names = ['stl', 'cnt', 'tv']
    self.metrics = L(LossMetrics(self.metric_names))

  def forward(self, pred, targ, fts, **kwargs):
    self.stl = self.stl_w*self.stl_loss_fn(pred, targ, fts)
    self.cnt = self.cnt_w*self.cnt_loss_fn(pred, targ, fts)
    self.tv  = self.tv_w *self.tv_loss_fn (pred, targ, fts)
    return self.stl+self.cnt+self.tv

# Cell
@funcs_kwargs
class SuperResLoss(Module):
  _methods=['pixel_loss','cnt_loss']
  def __init__(self, **kwargs):
    self.metrics = L(LossMetrics(['pixel','cnt']))

  def forward(self, pred, targ, fts,**kwargs):
    self.pixel = self.pixel_loss(pred,targ,fts,**kwargs)
    self.cnt   = self.cnt_loss  (pred,targ,fts,**kwargs)
    return self.pixel+self.cnt

  @delegates(pixel_loss)
  def pixel_loss(self,pred,targ,fts,**kwargs): return pixel_loss(pred,targ,fts,**kwargs)
  @delegates(content_loss)
  def cnt_loss(self,pred,targ,fts,**kwargs): return content_loss(pred,targ,fts,**kwargs)


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"TEST_IMAGE": "00_core.ipynb",
         "zip_safe": "00_core.ipynb",
         "wmean": "00_core.ipynb",
         "create": "00_core.ipynb",
         "IncrementalSplitter": "01_data.ipynb",
         "coco_stats": "01_data.ipynb",
         "TensorImageX": "01_data.ipynb",
         "PILImageX": "01_data.ipynb",
         "NormalizeX": "01_data.ipynb",
         "style_blocks": "01_data.ipynb",
         "NormalizeAll": "01_data.ipynb",
         "prepare_feat_model": "02_layer_features.ipynb",
         "get_layers": "02_layer_features.ipynb",
         "FeatModels": "02_layer_features.ipynb",
         "LayerFeats": "02_layer_features.ipynb",
         "gram": "03_loss.ipynb",
         "ContentLoss": "03_loss.ipynb",
         "content_loss": "03_loss.ipynb",
         "StyleLoss": "03_loss.ipynb",
         "TVLoss": "03_loss.ipynb",
         "PixelLoss": "03_loss.ipynb",
         "pixel_loss": "03_loss.ipynb",
         "FastStyleLoss": "03_loss.ipynb",
         "SuperResLoss": "03_loss.ipynb",
         "style_learner": "04_learner.ipynb",
         "transformer_net": "04_models.ipynb",
         "TransformerNet": "04_models.ipynb",
         "FeatsCallback": "06_callback.ipynb"}

modules = ["core.py",
           "data.py",
           "layer_features.py",
           "loss.py",
           "learner.py",
           "models.py",
           "callback.py"]

doc_url = "https://lgvaz.github.io/faststyle/"

git_url = "https://github.com/lgvaz/faststyle/tree/master/"

def custom_doc_links(name): return None
